@@ -1,5 +1,14 @@
 from abc import ABC, abstractmethod
-from typing import TYPE_CHECKING, Type, Optional, ClassVar, Union
+from typing import (
+    TYPE_CHECKING,
+    Type,
+    Optional,
+    ClassVar,
+    Union,
+    TypeVar,
+    Generic,
+    cast,
+)
 
 import attr
 import numpy as np
@@ -11,15 +20,17 @@ from corrscope.spectrum import SpectrumConfig, DummySpectrum, LogFreqSpectrum
 from corrscope.util import find, obj_name, iround
 from corrscope.utils.trigger_util import (
     get_period,
+    UNKNOWN_PERIOD,
     normalize_buffer,
     lerp,
     MIN_AMPLITUDE,
     abs_max,
 )
-from corrscope.utils.windows import leftpad, midpad, rightpad, gaussian_or_zero
-from corrscope.wave import FLOAT
+from corrscope.utils.windows import midpad, gaussian_or_zero
+from corrscope.wave_common import f32
 
 if TYPE_CHECKING:
+    import numpy.typing as npt
     from corrscope.wave import Wave
     from corrscope.renderer import RendererFrontend
 
@@ -37,6 +48,11 @@ class _TriggerConfig:
 class MainTriggerConfig(
     _TriggerConfig, KeywordAttrs, always_dump="edge_direction post_trigger post_radius"
 ):
+    if TYPE_CHECKING:
+
+        def __call__(self, wave: "Wave", *args, **kwargs) -> "MainTrigger":
+            return self.cls(wave, self, *args, **kwargs)
+
     # Must be 1 or -1.
     # MainTrigger.__init__() multiplies `wave.amplification *= edge_direction`.
     # get_trigger() should ignore `edge_direction` and look for rising edges.
@@ -61,7 +77,11 @@ class MainTriggerConfig(
 
 class PostTriggerConfig(_TriggerConfig, KeywordAttrs):
     parent: MainTriggerConfig = attr.ib(init=False)  # TODO Unused
-    pass
+
+    if TYPE_CHECKING:
+
+        def __call__(self, wave: "Wave", *args, **kwargs) -> "PostTrigger":
+            return self.cls(wave, self, *args, **kwargs)
 
 
 def register_trigger(config_t: Type[_TriggerConfig]):
@@ -76,7 +96,10 @@ def register_trigger(config_t: Type[_TriggerConfig]):
     return inner
 
 
-class _Trigger(ABC):
+result = TypeVar("result")
+
+
+class _Trigger(ABC, Generic[result]):
     def __init__(
         self,
         wave: "Wave",
@@ -90,7 +113,7 @@ class _Trigger(ABC):
         self.cfg = cfg
         self._wave = wave
 
-        # TODO rename tsamp to buffer_nsamp
+        # TODO rename _tsamp to _trigger_subsmp, tsamp to subsamp
         self._tsamp = tsamp
         self._stride = stride
         self._fps = fps
@@ -100,17 +123,22 @@ class _Trigger(ABC):
         self._wave_idx = wave_idx
 
         # Subsamples per second
-        self.subsmp_s = self._wave.smp_s / self._stride
+        self.subsmp_per_s = self._wave.smp_s / self._stride
 
-        frame_dur = 1 / fps
-        # Subsamples per frame
-        self._tsamp_frame = self.time2tsamp(frame_dur)
+        seconds_per_frame = 1 / fps
+        # Full samples per frame
+        self._smp_per_frame = self.seconds_to_samp(seconds_per_frame)
 
-    def time2tsamp(self, time: float) -> int:
-        return round(time * self.subsmp_s)
+    def seconds_to_samp(self, time: float) -> int:
+        return round(time * self._wave.smp_s)
 
     def custom_line(
-        self, name: str, data: np.ndarray, offset: bool, invert: bool = True
+        self,
+        name: str,
+        data: np.ndarray,
+        xs: np.ndarray,
+        absolute: bool,
+        invert: bool = True,
     ):
         """
         :param offset:
@@ -132,28 +160,23 @@ class _Trigger(ABC):
         if invert:
             data *= np.copysign(1, self._wave.amplification)
         self._renderer.update_custom_line(
-            name, self._wave_idx, self._stride, data, offset=offset
+            name, self._wave_idx, self._stride, data, xs, absolute
         )
 
-    def custom_vline(self, name: str, x: int, offset: bool):
+    def custom_vline(self, name: str, x: int, absolute: bool):
         """See above for `offset`."""
         if self._renderer is None:
             return
         self._renderer.update_vline(
-            name, self._wave_idx, self._stride, x, offset=offset
+            name, self._wave_idx, self._stride, x, absolute=absolute
         )
 
-    def offset_viewport(self, offset: int):
-        if self._renderer is None:
-            return
-        self._renderer.offset_viewport(self._wave_idx, offset)
-
     @abstractmethod
-    def get_trigger(self, index: int, cache: "PerFrameCache") -> int:
+    def get_trigger(self, index: int, cache: "PerFrameCache") -> result:
         """
         :param index: sample index
-        :param cache: Information shared across all stacked triggers,
-            May be mutated by function.
+        :param cache: Information shared across all stacked triggers. result_mean is
+            None. May be mutated by function.
         :return: new sample index, corresponding to rising edge
         """
         ...
@@ -163,7 +186,16 @@ class _Trigger(ABC):
         pass
 
 
-class MainTrigger(_Trigger, ABC):
+@attr.dataclass
+class TriggerResult:
+    # new sample index, corresponding to rising edge
+    result: int
+
+    # Estimated frequency in cycle/sec (Hertz). None if unknown.
+    freq_estimate: Optional[float]
+
+
+class MainTrigger(_Trigger[TriggerResult], ABC):
     cfg: MainTriggerConfig
     post: Optional["PostTrigger"]
 
@@ -195,7 +227,7 @@ class MainTrigger(_Trigger, ABC):
         pass
 
 
-class PostTrigger(_Trigger, ABC):
+class PostTrigger(_Trigger[int], ABC):
     """A post-processing trigger should have stride=1,
     and no more post triggers. This is subject to change."""
 
@@ -226,73 +258,87 @@ class PerFrameCache:
     # NOTE: period is a *non-subsampled* period.
     # The period of subsampled data must be multiplied by stride.
     period: Optional[int] = None
-    mean: Optional[float] = None
+
+    # The mean of the trigger input, smoothed by mean_responsiveness, used for regular
+    # and post triggering.
+    smoothed_mean: Optional[float] = None
+
+    # The mean of the wave around the trigger result. The same segment of the wave is
+    # also used to update the buffer.
+    result_mean: Optional[float] = None
 
 
 # CorrelationTrigger
 
 
-class LagPrevention(KeywordAttrs):
-    max_frames: float = 1
-    transition_frames: float = 0.25
-
-    def __attrs_post_init__(self):
-        validate_param(self, "max_frames", 0, 1)
-        validate_param(self, "transition_frames", 0, self.max_frames)
-
-
 class CorrelationTriggerConfig(
     MainTriggerConfig,
     always_dump="""
+    mean_responsiveness
+    slope_width
     pitch_tracking
-    slope_strength slope_width
     """
     # deprecated
     " buffer_falloff ",
 ):
-    # get_trigger()
-    # Edge/area finding
+    ## General parameters ##
+
+    # Whether to subtract the mean from each frame
+    mean_responsiveness: float = 1.0
+
+    # Sign amplification
     sign_strength: float = 0
-    edge_strength: float
 
-    # Slope detection
-    slope_strength: float = 0
-    slope_width: float = with_units("period", default=0.07)
+    # Maximum distance to move, in terms of trigger_ms/trigger_samp (not in GUI)
+    trigger_diameter: float = 0.5
 
-    # Correlation detection (meow~ =^_^=)
-    buffer_strength: float = 1
+    # Maximum distance to move, in terms of estimated wave period (not in GUI)
+    trigger_radius_periods: Optional[float] = 1.5
 
-    # Both data and buffer uses Gaussian windows. std = wave_period * falloff.
-    # get_trigger()
-    data_falloff: float = 1.5
-
-    # _update_buffer()
-    buffer_falloff: float = 0.5
-
-    # Maximum distance to move
-    trigger_diameter: Optional[float] = 0.5
+    ## Period/frequency estimation (not in GUI) ##
 
+    # Minimum pitch change to recalculate _prev_slope_finder (not in GUI)
     recalc_semitones: float = 1.0
-    lag_prevention: LagPrevention = attr.ib(factory=LagPrevention)
 
-    # _update_buffer
-    responsiveness: float
-
-    # Period/frequency estimation (not in GUI)
+    # (not in GUI)
     max_freq: float = with_units("Hz", default=4000)
 
-    # Pitch tracking = compute spectrum.
-    pitch_tracking: Optional["SpectrumConfig"] = None
+    ## Edge triggering ##
+
+    # Competes against buffer_strength.
+    edge_strength: float
+
+    slope_width: float = with_units("period", default=0.25)
+
+    ## Correlation ##
+
+    # Competes against edge_strength.
+    buffer_strength: float = 1
+
+    # How much to update the buffer *after* each frame.
+    responsiveness: float
+
+    # Standard deviation of buffer window, in terms of estimated wave period. Used by
+    # _update_buffer() *after* each frame. (not in GUI)
+    buffer_falloff: float = 0.5
+
+    # Below a specific correlation quality, discard the buffer entirely.
+    reset_below: float = 0
+
+    # Whether to compute a spectrum to rescale correlation buffer in response to pitch
+    # changes. (GUI only has a checkbox)
+    pitch_tracking: Optional[SpectrumConfig] = None
 
     # region Legacy Aliases
     trigger_strength = Alias("edge_strength")
     falloff_width = Alias("buffer_falloff")
+    data_falloff = Alias("trigger_radius_periods")
     # endregion
 
     def __attrs_post_init__(self) -> None:
         MainTriggerConfig.__attrs_post_init__(self)
 
-        validate_param(self, "slope_width", 0, 0.5)
+        # Don't validate slope_width.
 
         validate_param(self, "responsiveness", 0, 1)
         # TODO trigger_falloff >= 0
@@ -308,208 +354,274 @@ def validate_param(self, key: str, begin: float, end: float) -> None:
 @register_trigger(CorrelationTriggerConfig)
 class CorrelationTrigger(MainTrigger):
     """
-    Assume that if get_trigger(x) == x, then data[[x-1, x]] == [<0, >0].
-    - edge detectors [halfN = N//2] > 0.
-    - So wave.get_around(x)[N//2] > 0.
-    - So wave.get_around(x) = [x - N//2 : ...]
+    Correlation-based trigger. it's complicated.
 
-    test_trigger() checks that get_around() works properly, for even/odd N.
-    See Wave.get_around() docstring.
+    The data read from the input _wave (length A+B+_trigger_diameter) is longer than
+    the correlation buffer (length A+B), to ensure that when sliding the buffer
+    across the wave, no edge effects are encountered. This is inspired by
+    overlap-save convolution, but adapted to cross-correlation.
     """
 
     cfg: CorrelationTriggerConfig
 
+    A: int
+    """(const) Number of subsamples on buffer/step's left (negative) side."""
+    B: int
+    """(const) Number of subsamples on buffer/step's right (positive) side. Equals A.
+    buffer_nsamp = A+B
+    """
+
+    _trigger_diameter: int
+    """(const) Distance (in subsamples) between the smallest and largest positions we can
+    output on a given frame. Approximately equal to 0.5 * (A+B). """
+
+    _corr_buffer: "npt.NDArray[f32]"
+    """(mutable) [A+B] Amplitude"""
+
+    _prev_mean: float
+    _prev_period: Optional[int]
+    _prev_slope_finder: "Optional[npt.NDArray[f32]]"
+    """(mutable) [A+B] Amplitude"""
+    _prev_window: "npt.NDArray[f32]"
+    """(mutable) [A+B] Amplitude"""
+
+    _prev_trigger: int
+    _frames_since_spectrum: int
+    _spectrum_calc: DummySpectrum
+
     @property
-    def scfg(self) -> SpectrumConfig:
+    def scfg(self) -> Optional[SpectrumConfig]:
         return self.cfg.pitch_tracking
 
-    def __init__(self, *args, **kwargs):
-        """
-        Correlation-based trigger which looks at a window of `trigger_tsamp` samples.
-        it's complicated
-        """
-        super().__init__(*args, **kwargs)
-        self._buffer_nsamp = self._tsamp
+    def calc_buffer_std(self, period: float) -> float:
+        return period * self.cfg.buffer_falloff
 
-        # (const) Multiplied by each frame of input audio.
-        # Zeroes out all data older than 1 frame old.
-        self._lag_prevention_window = self._calc_lag_prevention()
-        assert self._lag_prevention_window.dtype == FLOAT
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.A = self.B = self._tsamp // 2
+        kernel_size = self.A + self.B
+        self._trigger_diameter = int(kernel_size * self.cfg.trigger_diameter)
 
         # (mutable) Correlated with data (for triggering).
         # Updated with tightly windowed old data at various pitches.
-        self._buffer = np.zeros(
-            self._buffer_nsamp, dtype=FLOAT
-        )  # type: np.ndarray[FLOAT]
-
-        # (const) Added to self._buffer. Nonzero if edge triggering is nonzero.
-        # Left half is -edge_strength, right half is +edge_strength.
-        # ASCII art: --._|‾'--
-        self._edge_finder = self._calc_step()
-        assert self._edge_finder.dtype == FLOAT
+        self._corr_buffer = np.zeros(kernel_size, dtype=f32)
 
+        self._prev_mean = 0.0
         # Will be overwritten on the first frame.
-        self._prev_period: Optional[int] = None
-        self._prev_window: Optional[np.ndarray] = None
-        self._prev_slope_finder: Optional[np.ndarray] = None
+        self._prev_period = None
+        self._prev_slope_finder = None
+        self._prev_window = np.zeros(kernel_size, f32)
 
-        self._prev_trigger: int = 0
+        self._prev_trigger = 0
 
         # (mutable) Log-scaled spectrum
-        self.frames_since_spectrum = 0
+        self._frames_since_spectrum = 0
 
         if self.scfg:
             self._spectrum_calc = LogFreqSpectrum(
-                scfg=self.scfg, subsmp_s=self.subsmp_s, dummy_data=self._buffer
+                scfg=self.scfg,
+                subsmp_per_s=self.subsmp_per_s,
+                dummy_data=self._corr_buffer,
             )
         else:
             self._spectrum_calc = DummySpectrum()
 
-    def _calc_lag_prevention(self) -> np.ndarray:
-        """Returns input-data window,
-        which zeroes out all data older than 1-ish frame old.
-        See https://github.com/corrscope/corrscope/wiki/Correlation-Trigger
-        """
-        N = self._buffer_nsamp
-        halfN = N // 2
-
-        # - Create a cosine taper of `width` <= 1 frame
-        # - Right-pad(value=1, len=1 frame)
-        # - Place in left half of N-sample buffer.
-
-        # To avoid cutting off data, use a narrow transition zone (invariant to stride).
-        lag_prevention = self.cfg.lag_prevention
-        tsamp_frame = self._tsamp_frame
-        transition_nsamp = round(tsamp_frame * lag_prevention.transition_frames)
-
-        # Left half of a Hann cosine taper
-        # Width (type=subsample) = min(frame * lag_prevention, 1 frame)
-        assert transition_nsamp <= tsamp_frame
-        width = transition_nsamp
-        taper = windows.hann(width * 2)[:width]
-
-        # Right-pad=1 taper to lag_prevention.max_frames long [t-#*f, t]
-        taper = rightpad(taper, iround(tsamp_frame * lag_prevention.max_frames))
-
-        # Left-pad=0 taper to left `halfN` of data_taper [t-halfN, t]
-        taper = leftpad(taper, halfN)
-
-        # Generate left half-taper to prevent correlating with 1-frame-old data.
-        # Right-pad=1 taper to [t-halfN, t-halfN+N]
-        # TODO switch to rightpad()? Does it return FLOAT or not?
-        data_taper = np.ones(N, dtype=FLOAT)
-        data_taper[:halfN] = np.minimum(data_taper[:halfN], taper)
-
-        return data_taper
-
-    def _calc_step(self) -> np.ndarray:
-        """Step function used for approximate edge triggering."""
-
-        # Increasing buffer_falloff (width of buffer)
-        # causes buffer to affect triggering, more than the step function.
-        # So we multiply edge_strength (step function height) by buffer_falloff.
-
-        cfg = self.cfg
-        edge_strength = cfg.edge_strength * cfg.buffer_falloff
-
-        N = self._buffer_nsamp
-        halfN = N // 2
-
-        step = np.empty(N, dtype=FLOAT)  # type: np.ndarray[FLOAT]
-        step[:halfN] = -edge_strength / 2
-        step[halfN:] = edge_strength / 2
-        step *= windows.gaussian(N, std=halfN / 3)
-        return step
-
     def _calc_slope_finder(self, period: float) -> np.ndarray:
         """Called whenever period changes substantially.
-        Returns a kernel to be correlated with input data,
-        to find positive slopes."""
-
-        N = self._buffer_nsamp
-        halfN = N // 2
-        slope_finder = np.zeros(N)
+        Returns a kernel to be correlated with input data to find positive slopes,
+        with length A+B."""
 
         cfg = self.cfg
-        slope_width = max(iround(cfg.slope_width * period), 1)
-        slope_strength = cfg.slope_strength * cfg.buffer_falloff
+        kernel_size = self.A + self.B
 
-        slope_finder[halfN - slope_width : halfN] = -slope_strength
-        slope_finder[halfN : halfN + slope_width] = slope_strength
+        # noinspection PyTypeChecker
+        slope_width: float = np.clip(cfg.slope_width * period, 1.0, self.A / 3)
+
+        # This is a fudge factor. Adjust it until it feels right.
+        slope_strength = cfg.edge_strength * 5
+        # slope_width is 1.0 or greater, so this doesn't divide by 0.
+
+        slope_finder = np.empty(kernel_size, dtype=f32)  # type: np.ndarray[f32]
+        slope_finder[: self.A] = -slope_strength / 2
+        slope_finder[self.A :] = slope_strength / 2
+
+        slope_finder *= windows.gaussian(kernel_size, std=slope_width)
         return slope_finder
 
     # end setup
 
     # begin per-frame
-    def get_trigger(self, index: int, cache: "PerFrameCache") -> int:
-        N = self._buffer_nsamp
+    def get_trigger(self, pos: int, cache: "PerFrameCache") -> TriggerResult:
         cfg = self.cfg
 
-        # Get data (1D, downmixed to mono)
         stride = self._stride
-        data = self._wave.get_around(index, N, stride)
+
+        # Convert sizes to full samples (not trigger subsamples) when indexing into
+        # _wave.
+
+        # _trigger_diameter is defined as inclusive. The length of find_peak()'s
+        # corr variable is (A + _trigger_diameter + B) - (A + B) + 1, or
+        # _trigger_diameter + 1. This gives us a possible triggering range of
+        # _trigger_diameter inclusive, which is what we want.
+        data_nsubsmp = self.A + self._trigger_diameter + self.B
+
+        trigger_begin = max(
+            pos - self._smp_per_frame, pos - self._trigger_diameter // 2
+        )
+        data_begin = trigger_begin - stride * self.A
+
+        # Get subsampled data (1D, downmixed to mono)
+        # [data_nsubsmp = A + _trigger_diameter + B] Amplitude
+        data = self._wave.get_padded(
+            data_begin, data_begin + stride * data_nsubsmp, stride
+        )
+        assert data.size == data_nsubsmp, (data.size, data_nsubsmp)
 
         if cfg.sign_strength != 0:
             signs = sign_times_peak(data)
             data += cfg.sign_strength * signs
 
-        # Remove mean from data
-        data -= np.add.reduce(data) / N
+        # Remove mean from data, if enabled.
+        mean = np.add.reduce(data) / data.size
+        period_data = data - mean
 
-        # Window data
-        period = get_period(data, self.subsmp_s, self.cfg.max_freq, self)
+        if cfg.mean_responsiveness:
+            self._prev_mean += cfg.mean_responsiveness * (mean - self._prev_mean)
+            if cfg.mean_responsiveness != 1:
+                data -= self._prev_mean
+            else:
+                data = period_data
+        cache.smoothed_mean = self._prev_mean
+
+        # Use period to recompute slope finder (if enabled) and restrict trigger
+        # diameter.
+        period = get_period(period_data, self.subsmp_per_s, cfg.max_freq, self)
         cache.period = period * stride
 
         semitones = self._is_window_invalid(period)
         # If pitch changed...
         if semitones:
-            # Gaussian window
-            period_symmetric_window = gaussian_or_zero(N, period * cfg.data_falloff)
-
-            # Left-sided falloff
-            lag_prevention_window = self._lag_prevention_window
-
-            # Both combined.
-            window = np.minimum(period_symmetric_window, lag_prevention_window)
-
-            # Slope finder
             slope_finder = self._calc_slope_finder(period)
 
-            data *= window
-
             # If pitch tracking enabled, rescale buffer to match data's pitch.
             if self.scfg and (data != 0).any():
                 # Mutates self._buffer.
                 self.spectrum_rescale_buffer(data)
 
             self._prev_period = period
-            self._prev_window = window
             self._prev_slope_finder = slope_finder
         else:
-            window = self._prev_window
-            slope_finder = self._prev_slope_finder
+            slope_finder = cast(np.ndarray, self._prev_slope_finder)
 
-            data *= window
+        corr_enabled = bool(cfg.buffer_strength) and bool(cfg.responsiveness)
 
-        prev_buffer: np.ndarray = self._buffer * self.cfg.buffer_strength
-        prev_buffer += self._edge_finder + slope_finder
+        # Buffer sizes:
+        # data_nsubsmp = A + _trigger_diameter + B
+        kernel_size = self.A + self.B
+        corr_nsamp = self._trigger_diameter + 1
+        assert corr_nsamp == data_nsubsmp - kernel_size + 1
 
-        # Calculate correlation
-        if self.cfg.trigger_diameter is not None:
-            radius = round(N * self.cfg.trigger_diameter / 2)
+        # Check if buffer still lines up well with data.
+        if corr_enabled:
+            # array[corr_nsamp] Amplitude
+            corr_quality = signal.correlate_valid(data, self._corr_buffer)
+            assert len(corr_quality) == corr_nsamp
+
+            if cfg.reset_below > 0:
+                peak_idx = np.argmax(corr_quality)
+                peak_quality = corr_quality[peak_idx]
+
+                data_slice = data[peak_idx : peak_idx + kernel_size]
+
+                # Keep in sync with _update_buffer()!
+                windowed_slice = data_slice - mean
+                normalize_buffer(windowed_slice)
+                windowed_slice *= self._prev_window
+                self_quality = np.add.reduce(data_slice * windowed_slice)
+
+                relative_quality = peak_quality / (self_quality + 0.001)
+                should_reset = relative_quality < cfg.reset_below
+                if should_reset:
+                    corr_quality[:] = 0
+                    self._corr_buffer[:] = 0
+                    corr_enabled = False
         else:
-            radius = None
+            corr_quality = np.zeros(corr_nsamp, f32)
 
-        trigger_score = correlate_data(data, prev_buffer, radius)
-        peak_offset = trigger_score.peak
-        trigger = index + (stride * peak_offset)
+        # array[A+B] Amplitude
+        corr_kernel = slope_finder
+        del slope_finder
+        if corr_enabled:
+            corr_kernel += self._corr_buffer * cfg.buffer_strength
+
+        # `corr[x]` = correlation of kernel placed at position `x` in data.
+        # `corr_kernel` is not allowed to move past the boundaries of `data`.
+        corr = signal.correlate_valid(data, corr_kernel)
+        assert len(corr) == corr_nsamp
+
+        peaks = corr_quality
+        del corr_quality
+        peaks *= cfg.buffer_strength
+
+        if cfg.edge_strength:
+            # I want a half-open cumsum, where edge_score[0] = 0, [1] = data[A], [2] =
+            # data[A] + data[A+1], etc. But cumsum is inclusive, which causes tests to
+            # fail. So subtract 1 from the input range.
+            edge_score = np.cumsum(data[self.A - 1 : len(data) - self.B])
+
+            # The optimal edge alignment is the *minimum* cumulative sum, so invert
+            # the cumsum so the minimum amplitude maps to the highest score.
+            edge_score *= -cfg.edge_strength
+            peaks += edge_score
+
+        # Don't pick peaks more than `period * trigger_radius_periods` away from the
+        # center.
+        if cfg.trigger_radius_periods and period != UNKNOWN_PERIOD:
+            trigger_radius = round(period * cfg.trigger_radius_periods)
+        else:
+            trigger_radius = None
+
+        def find_peak(
+            corr: np.ndarray, peaks: np.ndarray, radius: Optional[int]
+        ) -> int:
+            """If radius is set, the returned offset is limited to ±radius from the
+            center of correlation.
+            """
+            assert len(corr) == len(peaks) == corr_nsamp
+            # returns double, not single/f32
+            begin_offset = 0
+
+            if radius is not None:
+                Ncorr = len(corr)
+                mid = Ncorr // 2
+
+                left = max(mid - radius, 0)
+                right = min(mid + radius + 1, Ncorr)
+
+                corr = corr[left:right]
+                peaks = peaks[left:right]
+                begin_offset = left
+
+            min_corr = np.min(corr)
+
+            # Only permit local maxima. This fixes triggering errors where the edge
+            # of the allowed range has higher correlation than edges in-bounds,
+            # but isn't a rising edge itself (a local maximum of alignment).
+            corr[:-1][peaks[:-1] < peaks[1:]] = min_corr
+            corr[1:][peaks[1:] < peaks[:-1]] = min_corr
+            corr[0] = corr[-1] = min_corr
+
+            # Find optimal offset
+            peak_offset = np.argmax(corr) + begin_offset  # type: int
+            return peak_offset
+
+        # Find correlation peak.
+        peak_offset = find_peak(corr, peaks, trigger_radius)
+        trigger = trigger_begin + stride * (peak_offset)
 
         del data
 
         if self.post:
-            new_data = self._wave.get_around(trigger, N, stride)
-            cache.mean = np.add.reduce(new_data) / N
-
             # Apply post trigger (before updating correlation buffer)
             trigger = self.post.get_trigger(trigger, cache)
 
@@ -517,15 +629,18 @@ class CorrelationTrigger(MainTrigger):
         self._prev_trigger = trigger = max(trigger, self._prev_trigger)
 
         # Update correlation buffer (distinct from visible area)
-        aligned = self._wave.get_around(trigger, N, stride)
-        if cache.mean is None:
-            cache.mean = np.add.reduce(aligned) / N
+        aligned = self._wave.get_around(trigger, kernel_size, stride)
+        cache.result_mean = np.add.reduce(aligned) / kernel_size
         self._update_buffer(aligned, cache)
 
-        self.frames_since_spectrum += 1
+        self._frames_since_spectrum += 1
 
-        self.offset_viewport(peak_offset)
-        return trigger
+        # period: subsmp/cyc
+        freq_estimate = self.subsmp_per_s / period if period else None
+        # freq_estimate: cyc/s
+        # If period is 0 (unknown), freq_estimate is None.
+
+        return TriggerResult(trigger, freq_estimate)
 
     def spectrum_rescale_buffer(self, data: np.ndarray) -> None:
         """
@@ -535,10 +650,10 @@ class CorrelationTrigger(MainTrigger):
 
         # Setup
         scfg = self.scfg
-        N = self._buffer_nsamp
-        if self.frames_since_spectrum < self.scfg.min_frames_between_recompute:
+        Ntrigger = self._corr_buffer.size
+        if self._frames_since_spectrum < self.scfg.min_frames_between_recompute:
             return
-        self.frames_since_spectrum = 0
+        self._frames_since_spectrum = 0
 
         calc_spectrum = self._spectrum_calc.calc_spectrum
 
@@ -548,7 +663,7 @@ class CorrelationTrigger(MainTrigger):
         assert not np.any(np.isnan(spectrum))
 
         # Compute log-frequency spectrum of `self._buffer`.
-        prev_spectrum = calc_spectrum(self._buffer)
+        prev_spectrum = calc_spectrum(self._corr_buffer)
         # Don't normalize self._spectrum. It was already normalized when being assigned.
 
         # Rescale `self._buffer` until its pitch matches `data`.
@@ -557,18 +672,20 @@ class CorrelationTrigger(MainTrigger):
         ).peak
         if resample_notes != 0:
             # If we want to double pitch, we must divide data length by 2.
-            new_len = iround(N / 2 ** (resample_notes / scfg.notes_per_octave))
+            new_len = iround(Ntrigger / 2 ** (resample_notes / scfg.notes_per_octave))
 
-            def rescale_mut(in_buf):
+            def rescale_mut(corr_kernel_mut):
                 buf = np.interp(
-                    np.linspace(0, 1, new_len), np.linspace(0, 1, N), in_buf
+                    np.linspace(0, 1, new_len),
+                    np.linspace(0, 1, Ntrigger),
+                    corr_kernel_mut,
                 )
                 # assert len(buf) == new_len
-                buf = midpad(buf, N)
-                in_buf[:] = buf
+                buf = midpad(buf, Ntrigger)
+                corr_kernel_mut[:] = buf
 
             # Copy+resample self._buffer.
-            rescale_mut(self._buffer)
+            rescale_mut(self._corr_buffer)
 
     def _is_window_invalid(self, period: int) -> Union[bool, float]:
         """
@@ -607,44 +724,42 @@ class CorrelationTrigger(MainTrigger):
 
         :param data: Wave data. WILL BE MODIFIED.
         """
-        assert cache.mean is not None
+        assert cache.result_mean is not None
         assert cache.period is not None
-        buffer_falloff = self.cfg.buffer_falloff
         responsiveness = self.cfg.responsiveness
 
-        N = len(data)
-        if N != self._buffer_nsamp:
-            raise ValueError(
-                f"invalid data length {len(data)} does not match "
-                f"CorrelationTrigger {self._buffer_nsamp}"
+        if self.cfg.buffer_strength and responsiveness:
+            # N should equal self.A + self.B.
+            N = len(data)
+            if N != self._corr_buffer.size:
+                raise ValueError(
+                    f"invalid data length {len(data)} does not match "
+                    f"CorrelationTrigger {self._corr_buffer.size}"
+                )
+
+            # New waveform
+            data -= cache.result_mean
+            normalize_buffer(data)
+            window = gaussian_or_zero(
+                N, std=self.calc_buffer_std(cache.period / self._stride)
             )
+            data *= window
+            self._prev_window = window
 
-        # New waveform
-        data -= cache.mean
-        normalize_buffer(data)
-        window = gaussian_or_zero(N, std=(cache.period / self._stride) * buffer_falloff)
-        data *= window
-
-        # Old buffer
-        normalize_buffer(self._buffer)
-        self._buffer = lerp(self._buffer, data, responsiveness)
+            # Old buffer
+            normalize_buffer(self._corr_buffer)
+            self._corr_buffer = lerp(self._corr_buffer, data, responsiveness)
 
 
 @attr.dataclass
-class CorrelationResult:
+class SpectrumResult:
     peak: int
     corr: np.ndarray
 
 
-@attr.dataclass
-class InterpolatedCorrelationResult:
-    peak: float
-    corr: np.ndarray
-
-
-def correlate_data(
+def correlate_spectrum(
     data: np.ndarray, prev_buffer: np.ndarray, radius: Optional[int]
-) -> CorrelationResult:
+) -> SpectrumResult:
     """
     This is confusing.
 
@@ -656,9 +771,13 @@ def correlate_data(
     or we must slide data to the left (by sliding index to the right):
     - correlate(data, prev_buffer)
     - trigger = index + peak_offset
+
+    In correlate_spectrum(), I used to use parabolic() on the return value,
+    but unfortunately it was unreliable and caused Plok Beach bass to jitter,
+    so I turned it off (resulting in the same code as correlate_data).
     """
     N = len(data)
-    corr = signal.correlate(data, prev_buffer)  # returns double, not single/FLOAT
+    corr = signal.correlate(data, prev_buffer)  # returns double, not single/f32
     Ncorr = 2 * N - 1
     assert len(corr) == Ncorr
 
@@ -675,18 +794,7 @@ def correlate_data(
     # argmax(corr) == mid + peak_offset == (data >> peak_offset)
     # peak_offset == argmax(corr) - mid
     peak_offset = np.argmax(corr) - mid  # type: int
-    return CorrelationResult(peak_offset, corr)
-
-
-def correlate_spectrum(
-    data: np.ndarray, prev_buffer: np.ndarray, radius: Optional[int]
-) -> CorrelationResult:
-    """
-    I used to use parabolic() on the return value,
-    but unfortunately it was unreliable and caused Plok Beach bass to jitter,
-    so I turned it off (resulting in the same code as correlate_data).
-    """
-    return correlate_data(data, prev_buffer, radius)
+    return SpectrumResult(peak_offset, corr)
 
 
 def parabolic(xint: int, ys: np.ndarray) -> float:
@@ -746,7 +854,7 @@ class ZeroCrossingTrigger(PostTrigger):
     def get_trigger(self, index: int, cache: "PerFrameCache") -> int:
         radius = self._tsamp
 
-        wave = self._wave.with_offset(-cache.mean)
+        wave = self._wave.with_offset(-cache.smoothed_mean)
 
         if not 0 <= index < wave.nsamp:
             return index
@@ -796,5 +904,5 @@ class NullTriggerConfig(MainTriggerConfig):
 
 @register_trigger(NullTriggerConfig)
 class NullTrigger(MainTrigger):
-    def get_trigger(self, index: int, cache: "PerFrameCache") -> int:
-        return index
+    def get_trigger(self, index: int, cache: "PerFrameCache") -> TriggerResult:
+        return TriggerResult(index, None)
